"""Unifi implementation."""

import logging

from pprint import pformat

from aiohttp import client_exceptions

from .clients import Clients, URL as client_url, ClientsAll, URL_ALL as all_client_url
from .devices import Devices, URL as device_url
from .errors import raise_error, LoginRequired, ResponseError, RequestError
from .events import event
from .websocket import WSClient, SIGNAL_CONNECTION_STATE, SIGNAL_DATA
from .wlan import Wlans, URL as wlan_url

LOGGER = logging.getLogger(__name__)

MESSAGE_CLIENT = "sta:sync"
MESSAGE_CLIENT_REMOVED = "user:delete"
MESSAGE_DEVICE = "device:sync"
MESSAGE_EVENT = "events"

ATTR_MESSAGE = "message"
ATTR_META = "meta"
ATTR_DATA = "data"

DATA_CLIENT = "client"
DATA_CLIENT_REMOVED = "client_removed"
DATA_DEVICE = "device"
DATA_EVENT = "event"


class Controller:
    """Control a UniFi controller."""

    def __init__(
        self,
        host,
        websession,
        *,
        username,
        password,
        port=8443,
        site="default",
        sslcontext=None,
        callback=None,
    ):
        self.host = host
        self.session = websession
        self.port = port
        self.username = username
        self.password = password
        self.site = site
        self.sslcontext = sslcontext
        self.callback = callback

        self.url = f"https://{self.host}:{self.port}"
        self.is_unifi_os = False
        self.headers = None

        self.websocket = None

        self.clients = None
        self.clients_all = None
        self.devices = None
        self.wlans = None

    async def check_unifi_os(self):
        response = await self.request("get", url=self.url, allow_redirects=False)
        if response.status == 200:
            self.is_unifi_os = True
            self.headers = {"x-csrf-token": response.headers.get("x-csrf-token")}

    async def login(self):
        if self.is_unifi_os:
            url = f"{self.url}/api/auth/login"
        else:
            url = f"{self.url}/api/login"

        auth = {
            "username": self.username,
            "password": self.password,
            "remember": True,
        }

        await self.request("post", url=url, json=auth)

    async def sites(self):
        if self.is_unifi_os:
            url = f"{self.url}/proxy/network/api/self/sites"
        else:
            url = f"{self.url}/api/self/sites"

        sites = await self.request("get", url=url)
        LOGGER.debug(pformat(sites))
        return {site["desc"]: site for site in sites}

    async def initialize(self):
        clients = await self.request("get", client_url)
        self.clients = Clients(clients, self.request)
        devices = await self.request("get", device_url)
        self.devices = Devices(devices, self.request)
        all_clients = await self.request("get", all_client_url)
        self.clients_all = ClientsAll(all_clients, self.request)
        wlans = await self.request("get", wlan_url)
        self.wlans = Wlans(wlans, self.request)

    def start_websocket(self):
        """Start websession and websocket to UniFi."""
        self.websocket = WSClient(
            self.session,
            self.host,
            self.port,
            self.sslcontext,
            self.site,
            callback=self.session_handler,
            is_unifi_os=self.is_unifi_os,
        )
        self.websocket.start()

    def stop_websocket(self) -> None:
        """Close websession and websocket to UniFi."""
        LOGGER.info("Shutting down connections to UniFi.")
        if self.websocket:
            self.websocket.stop()

    def session_handler(self, signal: str) -> None:
        """Signalling from websocket.

           data - new data available for processing.
           state - network state has changed.
        """
        if not self.websocket:
            return

        if signal == SIGNAL_DATA:
            new_items = self.message_handler(self.websocket.data)
            if new_items and self.callback:
                self.callback(SIGNAL_DATA, new_items)

        elif signal == SIGNAL_CONNECTION_STATE and self.callback:
            self.callback(SIGNAL_CONNECTION_STATE, self.websocket.state)

    def message_handler(self, message: dict) -> dict:
        """Receive event from websocket and identifies where the event belong."""
        changes = {}

        if message[ATTR_META][ATTR_MESSAGE] == MESSAGE_CLIENT:
            changes[DATA_CLIENT] = self.clients.process_raw(message[ATTR_DATA])

        elif message[ATTR_META][ATTR_MESSAGE] == MESSAGE_DEVICE:
            changes[DATA_DEVICE] = self.devices.process_raw(message[ATTR_DATA])

        elif message[ATTR_META][ATTR_MESSAGE] == MESSAGE_EVENT:
            events = []
            for item in message[ATTR_DATA]:
                events.append(event(item))
            self.clients.process_event(events)
            changes[DATA_EVENT] = set(events)

        elif message[ATTR_META][ATTR_MESSAGE] == MESSAGE_CLIENT_REMOVED:
            changes[DATA_CLIENT_REMOVED] = self.clients.remove(message[ATTR_DATA])

        else:
            LOGGER.debug(f"Unsupported message type {message}")

        return changes

    async def request(self, method, path=None, json=None, url=None, **kwargs):
        """Make a request to the API."""
        if not url:
            if self.is_unifi_os:
                url = f"{self.url}/proxy/network/api/s/{self.site}"
            else:
                url = f"{self.url}/api/s/{self.site}"

            if path is not None:
                url += f"{path}"

        LOGGER.debug("%s", url)

        try:
            async with self.session.request(
                method,
                url,
                json=json,
                ssl=self.sslcontext,
                headers=self.headers,
                **kwargs,
            ) as res:
                LOGGER.debug("%s %s %s", res.status, res.content_type, res)

                if res.status == 401:
                    raise LoginRequired(f"Call {url} received 401 Unauthorized")

                if res.status == 404:
                    raise ResponseError(f"Call {url} received 404 Not Found")

                if res.content_type == "application/json":
                    response = await res.json()
                    _raise_on_error(response)
                    if "data" in response:
                        return response["data"]
                    return response
                return res

        except client_exceptions.ClientError as err:
            raise RequestError(
                f"Error requesting data from {self.host}: {err}"
            ) from None


def _raise_on_error(data):
    """Check response for error message."""
    if not isinstance(data, dict):
        return

    if "meta" in data and data["meta"]["rc"] == "error":
        raise_error(data["meta"]["msg"])

    if "errors" in data:
        raise_error(data["errors"][0])



"""Clients are devices on a UniFi network."""

from .api import APIItem, APIItems

URL = "/stat/sta"  # Active clients
URL_ALL = "/rest/user"  # All known and configured clients

URL_CLIENT_STATE_MANAGER = "/cmd/stamgr"


class Clients(APIItems):
    """Represents client network devices."""

    KEY = "mac"

    def __init__(self, raw, request):
        super().__init__(raw, request, URL, Client)

    async def async_block(self, mac):
        """Block client from controller."""
        data = {"mac": mac, "cmd": "block-sta"}
        await self._request("post", URL_CLIENT_STATE_MANAGER, json=data)

    async def async_unblock(self, mac):
        """Unblock client from controller."""
        data = {"mac": mac, "cmd": "unblock-sta"}
        await self._request("post", URL_CLIENT_STATE_MANAGER, json=data)


class ClientsAll(APIItems):
    """Represents all client network devices."""

    KEY = "mac"

    def __init__(self, raw, request):
        super().__init__(raw, request, URL_ALL, Client)


class Client(APIItem):
    """Represents a client network device."""

    @property
    def blocked(self):
        return self.raw.get("blocked", "")

    @property
    def essid(self):
        return self.raw.get("essid", "")

    @property
    def hostname(self):
        return self.raw.get("hostname", "")

    @property
    def ip(self):
        return self.raw.get("ip", "")

    @property
    def is_guest(self):
        return self.raw.get("is_guest", "")

    @property
    def is_wired(self):
        return self.raw.get("is_wired", "")

    @property
    def last_seen(self):
        return self.raw.get("last_seen", "")

    @property
    def mac(self):
        return self.raw.get("mac", "")

    @property
    def name(self):
        return self.raw.get("name", "")

    @property
    def oui(self):
        return self.raw.get("oui", "")

    @property
    def site_id(self):
        return self.raw.get("site_id", "")

    @property
    def sw_depth(self):
        return self.raw.get("sw_depth", "")

    @property
    def sw_mac(self):
        """MAC for switch client is connected to."""
        return self.raw.get("sw_mac", "")

    @property
    def sw_port(self):
        """Switch port client is connected to."""
        return self.raw.get("sw_port", "")

    @property
    def rx_bytes(self):
        """Bytes received over wireless connection."""
        return self.raw.get("rx_bytes", 0)

    @property
    def tx_bytes(self):
        """Bytes transferred over wireless connection."""
        return self.raw.get("tx_bytes", 0)

    @property
    def uptime(self):
        return self.raw.get("uptime", 0)

    @property
    def wired_rx_bytes(self):
        """Bytes received over wired connection."""
        return self.raw.get("wired-rx_bytes", 0)

    @property
    def wired_tx_bytes(self):
        """Bytes transferred over wired connection."""
        return self.raw.get("wired-tx_bytes", 0)

    def __repr__(self):
        """Return the representation."""
        name = self.name or self.hostname
        return f"<Client {name}: {self.mac} {self.raw}>"



"""WLANs as part of a UniFi network."""

from .api import APIItem, APIItems

URL = "/rest/wlanconf"  # List WLAN configuration


class Wlans(APIItems):
    """Represents WLAN configurations."""

    KEY = "name"

    def __init__(self, raw, request):
        super().__init__(raw, request, URL, Wlan)

    async def async_enable(self, wlan):
        """Block client from controller."""
        wlan_url = f"{URL}/{wlan.id}"
        data = {"enabled": True}
        await self._request("put", wlan_url, json=data)

    async def async_disable(self, wlan):
        """Unblock client from controller."""
        wlan_url = f"{URL}/{wlan.id}"
        data = {"enabled": False}
        await self._request("put", wlan_url, json=data)


class Wlan(APIItem):
    """Represents a WLAN configuration."""

    @property
    def id(self):
        return self.raw["_id"]

    @property
    def bc_filter_enabled(self):
        return self.raw.get("bc_filter_enabled", False)

    @property
    def bc_filter_list(self):
        return self.raw["bc_filter_list"]

    @property
    def dtim_mode(self):
        return self.raw["dtim_mode"]

    @property
    def dtim_na(self):
        return self.raw["dtim_na"]

    @property
    def dtim_ng(self):
        return self.raw["dtim_ng"]

    @property
    def enabled(self):
        return self.raw["enabled"]

    @property
    def group_rekey(self):
        return self.raw["group_rekey"]

    @property
    def is_guest(self):
        return self.raw.get("is_guest", False)

    @property
    def mac_filter_enabled(self):
        return self.raw.get("mac_filter_enabled", False)

    @property
    def mac_filter_list(self):
        return self.raw["mac_filter_list"]

    @property
    def mac_filter_policy(self):
        return self.raw["mac_filter_policy"]

    @property
    def minrate_na_advertising_rates(self):
        return self.raw["minrate_na_advertising_rates"]

    @property
    def minrate_na_beacon_rate_kbps(self):
        return self.raw["minrate_na_beacon_rate_kbps"]

    @property
    def minrate_na_data_rate_kbps(self):
        return self.raw["minrate_na_data_rate_kbps"]

    @property
    def minrate_na_enabled(self):
        return self.raw.get("minrate_na_enabled", False)

    @property
    def minrate_na_mgmt_rate_kbps(self):
        return self.raw["minrate_na_mgmt_rate_kbps"]

    @property
    def minrate_ng_advertising_rates(self):
        return self.raw["minrate_ng_advertising_rates"]

    @property
    def minrate_ng_beacon_rate_kbps(self):
        return self.raw["minrate_ng_beacon_rate_kbps"]

    @property
    def minrate_ng_cck_rates_enabled(self):
        return self.raw.get("minrate_ng_cck_rates_enabled", False)

    @property
    def minrate_ng_data_rate_kbps(self):
        return self.raw["minrate_ng_data_rate_kbps"]

    @property
    def minrate_ng_enabled(self):
        return self.raw.get("minrate_ng_enabled", False)

    @property
    def minrate_ng_mgmt_rate_kbps(self):
        return self.raw["minrate_ng_mgmt_rate_kbps"]

    @property
    def name(self):
        return self.raw["name"]

    @property
    def no2ghz_oui(self):
        return self.raw["no2ghz_oui"]

    @property
    def schedule(self):
        return self.raw["schedule"]

    @property
    def security(self):
        return self.raw["security"]

    @property
    def site_id(self):
        return self.raw["site_id"]

    @property
    def usergroup_id(self):
        return self.raw["usergroup_id"]

    @property
    def wep_idx(self):
        return self.raw["wep_idx"]

    @property
    def wlangroup_id(self):
        return self.raw["wlangroup_id"]

    @property
    def wpa_enc(self):
        return self.raw["wpa_enc"]

    @property
    def wpa_mode(self):
        return self.raw["wpa_mode"]

    @property
    def x_iapp_key(self):
        return self.raw["x_iapp_key"]

    @property
    def x_passphrase(self):
        return self.raw["x_passphrase"]



"""UniFi devices are network infrastructure.

Access points, Gateways, Switches.
"""

import logging

from .api import APIItem, APIItems

LOGGER = logging.getLogger(__name__)

URL = "/stat/device"


class Devices(APIItems):
    """Represents network devices."""

    KEY = "mac"

    def __init__(self, raw, request):
        super().__init__(raw, request, URL, Device)


class Device(APIItem):
    """Represents a network device."""

    def __init__(self, raw, request):
        super().__init__(raw, request)
        self.ports = Ports(raw.get("port_table", []))

    def update(self, raw):
        self.ports.update(raw.get("port_table", []))
        super().update(raw)

    @property
    def board_rev(self):
        return self.raw["board_rev"]

    @property
    def disabled(self):
        return self.raw.get("disabled", False)

    @property
    def id(self):
        return self.raw["device_id"]

    @property
    def ip(self):
        return self.raw["ip"]

    @property
    def fan_level(self):
        return self.raw.get("fan_level")

    @property
    def has_fan(self):
        return self.raw.get("has_fan", False)

    @property
    def last_seen(self):
        return self.raw.get("last_seen")

    @property
    def mac(self):
        return self.raw["mac"]

    @property
    def model(self):
        return self.raw["model"]

    @property
    def name(self):
        return self.raw.get("name")

    @property
    def overheating(self):
        return self.raw.get("overheating")

    @property
    def port_overrides(self):
        return self.raw.get("port_overrides", [])

    @property
    def port_table(self):
        return self.raw.get("port_table", [])

    @property
    def state(self):
        return self.raw["state"]

    @property
    def type(self):
        return self.raw["type"]

    @property
    def version(self):
        return self.raw["version"]

    @property
    def upgradable(self):
        return self.raw.get("upgradable")

    async def async_set_port_poe_mode(self, port_idx, mode):
        """Set port poe mode.

        Auto, 24v, passthrough, off.
        Make sure to not overwrite any existing configs.
        """
        LOGGER.debug("Override port %d with mode %s", port_idx, mode)

        no_existing_config = True
        for port_override in self.port_overrides:
            if port_idx == port_override["port_idx"]:
                port_override["poe_mode"] = mode
                no_existing_config = False
                break

        if no_existing_config:
            self.port_overrides.append(
                {
                    "port_idx": port_idx,
                    "portconf_id": self.ports[port_idx].portconf_id,
                    "poe_mode": mode,
                }
            )

        url = f"/rest/device/{self.id}"
        data = {"port_overrides": self.port_overrides}

        await self._request("put", url, json=data)

    def __repr__(self):
        """Return the representation."""
        return f"<Device {self.name}: {self.mac}>"


class Ports:
    """Represents ports on a device."""

    def __init__(self, raw_list):
        self.ports = {}
        for raw in raw_list:
            port = Port(raw)
            index = None
            if port.port_idx is not None:
                index = port.port_idx
            elif port.ifname is not None:
                index = port.ifname
            if index is not None:
                self.ports[index] = port

    def update(self, raw_list):
        for raw in raw_list:
            index = None

            if "port_idx" in raw:
                index = raw["port_idx"]

            elif "ifname" in raw:
                index = raw["ifname"]

            if index in self.ports:
                self.ports[index].raw = raw

    def values(self):
        return self.ports.values()

    def __getitem__(self, obj_id):
        return self.ports[obj_id]

    def __iter__(self):
        return iter(self.ports)


class Port:
    """Represents a network port."""

    def __init__(self, raw):
        self.raw = raw

    @property
    def ifname(self):
        """Used by USG."""
        return self.raw.get("ifname")

    @property
    def media(self):
        return self.raw.get("media")

    @property
    def name(self):
        return self.raw["name"]

    @property
    def port_idx(self):
        return self.raw.get("port_idx")

    @property
    def poe_class(self):
        return self.raw.get("poe_class")

    @property
    def poe_enable(self):
        """Indicates if Poe is supported/requested by client."""
        return self.raw.get("poe_enable")

    @property
    def poe_mode(self):
        """Indicates if Poe is auto, pasv24, passthrough, off or None"""
        return self.raw.get("poe_mode")

    @property
    def poe_power(self):
        return self.raw.get("poe_power")

    @property
    def poe_voltage(self):
        return self.raw.get("poe_voltage")

    @property
    def portconf_id(self):
        return self.raw.get("portconf_id")

    @property
    def port_poe(self):
        return self.raw.get("port_poe") is True

    @property
    def up(self):
        return self.raw.get("up")

    def __repr__(self):
        """Return the representation."""
        return f"<{self.name}: Poe {self.poe_enable}>"



"""API management class and base class for the different end points."""

import logging

from pprint import pformat

LOGGER = logging.getLogger(__name__)

SOURCE_DATA = "data"
SOURCE_EVENT = "event"


class APIItems:
    """Base class for a map of API Items."""

    KEY = None

    def __init__(self, raw, request, path, item_cls):
        self._request = request
        self._path = path
        self._item_cls = item_cls
        self._items = {}
        self.process_raw(raw)
        LOGGER.debug(pformat(raw))

    async def update(self):
        raw = await self._request("get", self._path)
        self.process_raw(raw)

    def process_raw(self, raw: list) -> set:
        new_items = set()

        for raw_item in raw:
            key = raw_item[self.KEY]
            obj = self._items.get(key)

            if obj is not None:
                obj.update(raw=raw_item)
            else:
                self._items[key] = self._item_cls(raw_item, self._request)
                new_items.add(key)

        return new_items

    def process_event(self, events: list) -> set:
        new_items = set()

        for event in events:
            obj = self._items.get(event.mac)

            if obj is not None:
                obj.update(event=event)
                new_items.add(event.mac)

        return new_items

    def remove(self, raw: list) -> set:
        removed_items = set()

        for raw_item in raw:
            key = raw_item[self.KEY]

            if key in self._items:
                self._items.pop(key)
                removed_items.add(key)

        return removed_items

    def values(self):
        return self._items.values()

    def __getitem__(self, obj_id):
        try:
            return self._items[obj_id]
        except KeyError:
            LOGGER.error(f"Couldn't find key: {obj_id}")

    def __iter__(self):
        return iter(self._items)


class APIItem:
    """Base class for all end points using APIItems class."""

    def __init__(self, raw, request):
        self._raw = raw
        self._request = request
        self._event = None
        self._source = SOURCE_DATA
        self._callbacks = []

    @property
    def raw(self) -> dict:
        """Read only raw data."""
        return self._raw

    @property
    def event(self) -> dict:
        """Read only event data."""
        return self._event

    @property
    def last_updated(self) -> str:
        """Which source, data or event last called update."""
        return self._source

    def update(self, raw=None, event=None) -> None:
        """Update raw data and signal new data is available."""
        if raw:
            self._raw = raw
            self._source = SOURCE_DATA

        elif event:
            self._event = event
            self._source = SOURCE_EVENT

        else:
            return

        for signal_update in self._callbacks:
            signal_update()

    def register_callback(self, callback) -> None:
        """Register callback for signalling.

        Callback will be used by update.
        """
        self._callbacks.append(callback)

    def remove_callback(self, callback) -> None:
        """Remove registered callback."""
        if callback in self._callbacks:
            self._callbacks.remove(callback)



"""Setup for aioUnifi"""

# https://ubntwiki.com/products/software/unifi-controller/api
# https://jeffknupp.com/blog/2013/08/16/open-sourcing-a-python-project-the-right-way/
# http://peterdowns.com/posts/first-time-with-pypi.html
# pip install -e .
# Upload to PyPI Live
# python setup.py sdist bdist_wheel
# twine upload dist/aiounifi-* --skip-existing

from setuptools import setup

setup(
    name="aiounifi",
    packages=["aiounifi"],
    version="19",
    description="An asynchronous Python library for communicating with Unifi Controller API",
    author="Robert Svensson",
    author_email="Kane610@users.noreply.github.com",
    license="MIT",
    url="https://github.com/Kane610/aiounifi",
    download_url="https://github.com/Kane610/aiounifi/archive/v19.tar.gz",
    install_requires=["aiohttp"],
    keywords=["unifi", "homeassistant"],
    classifiers=["Natural Language :: English", "Programming Language :: Python :: 3"],
)
